import math
from collections import Counter
from typing import Dict, List, Set, Tuple, Union

import numpy as np
from scipy.stats.mstats import gmean
from spacy.tokens import Doc, Span, SpanGroup, Token


FrequencyDist = Dict[str, float]
SpanBoundaries = Dict[str, List[str]]


def weighted_average(
    span_metrics: Dict[str, Dict[str, float]],
    frequencies: Dict[str, Counter],
) -> Dict[str, float]:
    weighted_average = {}
    for spans_key, span_type_dict in span_metrics.items():
        weighted_average[spans_key] = np.average(
            list(span_type_dict.values()),
            weights=list(frequencies[spans_key].values()),
        )

    return weighted_average


class SpanAnalyzer:
    def __init__(self, docs: List[Doc]):
        self.docs = docs
        self.p_corpus = self._get_distribution(self.docs, normalize=True)
        self.keys = self._get_all_keys()

    @property
    def frequency(self) -> Dict[str, Counter]:
        """Number of spans for a span type in the dataset's training corpus.

        Frequency tends to be positively correlated with performance. However,
        architectural choices (like the use of transfer learning) can reduce
        data requirements for ML models. For such architectures, there is a
        small correlation between frequency and performance.
        """
        frequency = {}
        for doc in self.docs:
            for spans_key in list(doc.spans.keys()):
                if spans_key not in frequency:
                    frequency[spans_key] = Counter()
                for span in doc.spans[spans_key]:
                    if span.label_ is None:
                        continue
                    else:
                        frequency[spans_key][span.label_] += 1
        return frequency

    @property
    def length(self) -> Dict[str, Dict[str, float]]:
        """Geometric mean of the spans' lengths in tokens.

        Traditional CRF models tend to perform poorly at the identification of
        long spans due to their strict Markov assumption. Architectures that
        rely on such assumptions should follow the same pattern. On the other
        hand, LSTMs or Transformers should do better on long spans.
        """
        _length = {}
        for doc in self.docs:
            for spans_key in list(doc.spans.keys()):
                if spans_key not in _length:
                    _length[spans_key] = {}
                for span in doc.spans[spans_key]:
                    if span.label_ is None:
                        continue
                    elif span.label_ not in _length[spans_key]:
                        _length[spans_key][span.label_] = []
                    else:
                        _length[spans_key][span.label_].append(len(span))

        # Compute for the geometric mean for each span type
        length = {}
        for spans_key, span_type_dict in _length.items():
            length[spans_key] = {}
            for span_type, lengths in span_type_dict.items():
                length[spans_key][span_type] = gmean(lengths)

        return length

    @property
    def span_distinctiveness(self) -> Dict[str, float]:
        """Distinctiveness of the span compared to the corpus.

        Measures how distinct the text comprising the spans compared to the
        rest of the corpus. It is defined as the KL divergence D(P_span || P),
        where P is the unigram word distribution of the corpus, and P_span as
        the unigram distribution of tokens within the span.

        High values indicate that different words are used inside spans
        compared to the rest of the text, whereas low values indicate that the
        word distribution is similar inside and outside of spans.

        This property is positively correlated with model performance. Spans
        with high distinctiveness should be able to rely more heavily on local
        features, as each token carries information about span membership. Low
        span distrinctivess then calls for sequence information.
        """
        p_spans: Dict[str, Dict[str, FrequencyDist]] = {}
        for key in self.keys:
            p_spans[key] = {}
            spans_per_type = self._get_all_spans_in_key(key)
            for span_type, spans in spans_per_type.items():
                p_span = self._get_distribution(spans, normalize=True)
                p_spans[key][span_type] = p_span

        # Compute value for each spans_key
        span_distincts: Dict[str, Dict[str, float]] = {}
        for spans_key, freq_per_type in p_spans.items():
            span_distincts[spans_key] = {}
            for span_type, frequency_dist in freq_per_type.items():
                span_distincts[spans_key][span_type] = self._get_kl_divergence(
                    frequency_dist, self.p_corpus
                )
        return span_distincts

    @property
    def boundary_distinctiveness(self):
        """Distinctiveness of the boundaries compared to the corpus.

        Measures how distinctive the starts and ends of spans are. It is
        formalized as the KL-divergence D(P_bounds || P) where P is the unigram
        word distribution of the corpus, and P_bounds as the unigram
        distribution of the boundary tokens.

        This property is positively correlated with model performance. High
        values mean that the start and end points of spans are easy to spot,
        while low values indicate smooth transitions.
        """
        p_bounds: Dict[str, Dict[str, FrequencyDist]] = {}
        for key in self.keys:
            p_bounds[key] = {}
            span_boundaries_per_type = self._get_all_boundaries_in_key(key)
            for span_type, span_bounds in span_boundaries_per_type.items():
                bounds = span_bounds["start"] + span_bounds["end"]
                p_bound = self._get_distribution(bounds, normalize=True, unigrams=True)
                p_bounds[key][span_type] = p_bound

        # Compute value for each spans_key
        bound_distincts: Dict[str, Dict[str, float]] = {}
        for spans_key, freq_per_type in p_bounds.items():
            bound_distincts[spans_key] = {}
            for span_type, frequency_dist in freq_per_type.items():
                bound_distincts[spans_key][span_type] = self._get_kl_divergence(
                    frequency_dist, self.p_corpus
                )

        return bound_distincts

    def _get_all_keys(self) -> Set[str]:
        """Get all spans_key in the corpus"""
        return set([key for doc in self.docs for key in list(doc.spans.keys())])

    def _get_all_spans_in_key(self, spans_key: str) -> Dict[str, List[Span]]:
        """Get all spans given a specified spans_key

        Returns a dictionary where the keys are the span types and
        the values are the list of Spans within that span type.
        """
        spans: Dict[str, List[Span]] = {}
        for doc in self.docs:
            for span in doc.spans[spans_key]:
                if span.label_ not in spans:
                    spans[span.label_] = []
                else:
                    spans[span.label_].append(span)

        return spans

    def _get_all_boundaries_in_key(self, spans_key: str) -> Dict[str, SpanBoundaries]:
        """Get the boundary tokens for all spans in a spans_key

        Returns a dictionary where the keys are the span types and
        the values are a tuple of lists for the start and end span
        boundaries.
        """
        bounds: Dict[str, SpanBoundaries] = {}
        for doc in self.docs:
            for span in doc.spans[spans_key]:
                if span.label_ not in bounds:
                    bounds[span.label_] = {"start": [], "end": []}
                else:
                    # Get span boundaries
                    span_bound_start_idx = span.start - 1
                    if span_bound_start_idx >= 0:
                        bounds[span.label_]["start"].append(doc[span_bound_start_idx])

                    span_bound_end_idx = span.end + 1
                    if span_bound_end_idx < len(doc):
                        bounds[span.label_]["end"].append(doc[span_bound_end_idx])
        return bounds

    def _get_distribution(
        self,
        texts: Union[List[Doc], List[SpanGroup], List[Token]],
        normalize: bool = True,
        unigrams: bool = False,
    ) -> Counter:
        """Get each word's sample frequency given a list of text"""
        word_counts = Counter()
        for text in texts:
            if unigrams:
                word_counts[self._normalize_text(text.text)] += 1
            else:
                for token in text:
                    word_counts[self._normalize_text(token.text)] += 1

        if normalize:
            total = sum(word_counts.values(), 0.0)
            word_counts = Counter({k: v / total for k, v in word_counts.items()})

        return word_counts

    def _get_kl_divergence(self, p: Counter, q: Counter) -> float:
        """Compute the Kullback-Leibler divergence

        The parameters p and q are taken as unigram word distributions
        """
        total = 0
        for word, p_word in p.items():
            total += p_word * math.log(p_word / q[word])
        return total

    def _normalize_text(self, text: str) -> str:
        # The source code from the paper converts all digits into a single
        # value, '0'. Perhaps this is a way to "normalize" all numerical value
        # I'm not sure if this is generalizable so I'll comment it out for the meantime.

        # pattern = re.compile("[0-9]")
        # text = re.sub(pattern, "0", text)
        text = text.lower().replace("``", '"').replace("''", '"')
        return text



import tarfile
from pathlib import Path
from typing import List, Union, Set

import requests
import spacy
import typer
from spacy.tokens import Doc, DocBin, Span, SpanGroup
from spacy.training.converters import conll_ner_to_docs
from tqdm import tqdm
from wasabi import msg

from .constants import ASSETS_PATH, CORPUS_PATH

URL = "https://www.ims.uni-stuttgart.de/documents/ressourcen/korpora/riqua/riqua.tar.gz"
FILENAME = "riqua.tar.gz"
ANNOTATIONS = ASSETS_PATH / "riqua" / "merged"


def download_riqua(url: str) -> Path:
    save_path = ASSETS_PATH / FILENAME
    resp = requests.get(url, stream=True)
    total = int(resp.headers.get("Content-Length", 0))

    with open(save_path, "wb") as file:
        pbar = tqdm(unit="B", total=total)
        for chunk in resp.iter_content(chunk_size=1024):
            if chunk:
                pbar.update(len(chunk))
                file.write(chunk)
    return save_path


def _convert_to_doc(
    file: str, nlp: spacy.language.Language, spans_key: str = "sc"
) -> Doc:
    """Convert BRAT into spaCy

    Entities are saved into doc.spans, not doc.ents for the purposes
    of this reproducibility study.
    """
    text_file = ANNOTATIONS / f"{file}.txt"
    annot_file = ANNOTATIONS / f"{file}.ann"

    with open(text_file, "r") as f:
        text_str = f.read()

    annotations = []
    with open(annot_file, "r") as f:
        for line in f:
            entry = line.split()
            if entry[0].startswith("T"):
                data = {
                    "label": entry[1],
                    "start": int(entry[2]),
                    "end": int(entry[3]),
                    "text": " ".join(entry[4:]),
                }
                annotations.append(data)

    # Customize the tokenizer so it splits on double-hyphen em-dashes
    suffixes = nlp.Defaults.suffixes + ["--"]
    suffix_regex = spacy.util.compile_suffix_regex(suffixes)
    nlp.tokenizer.suffix_search = suffix_regex.search

    doc = nlp(text_str)
    spans = []

    for annot in annotations:
        # FIXME: Maybe there's better alignment rules here
        # I'll just set the mode to expand just to see.
        span = doc.char_span(
            annot["start"],
            annot["end"],
            annot["label"],
            alignment_mode="expand",
        )
        if span is None:
            msg.warn(f"Found an empty span in {annot_file}: {annot}")
        spans.append(span)

    doc.spans[spans_key] = spans
    return doc


def parse_riqua(files: Union[List[str], Set[str]]) -> List[Doc]:
    docs = []
    nlp = spacy.blank("en")
    for file in tqdm(files):
        doc = _convert_to_doc(file, nlp)
        docs.append(doc)
    return docs


def main(
    url: str = typer.Option(default=URL, show_default=True),
    skip_download: bool = typer.Option(False, "--skip-download"),
):
    msg.info(f"Downloading files and extracting them")
    filepath = download_riqua(url) if not skip_download else ASSETS_PATH / FILENAME
    with tarfile.open(filepath) as f:
        for member in tqdm(f.getmembers(), total=len(f.getmembers())):
            f.extract(member, path=ASSETS_PATH)

    msg.info(f"Converting annotations into spaCy docs")
    files = set([f.stem for f in ANNOTATIONS.glob("**/*") if f.is_file()])
    if files:
        docs = parse_riqua(files)
    else:
        msg.fail(f"No files found in {ANNOTATIONS}")

    msg.info(f"Saving into DocBin")
    doc_bin = DocBin(docs=docs)
    doc_bin.to_disk(CORPUS_PATH / "riqua.spacy")
    msg.good(f"Saved to {CORPUS_PATH}")


if __name__ == "__main__":
    typer.run(main)



from inspect import cleandoc
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import spacy
import typer
from spacy.tokens import Doc, DocBin
from wasabi import msg

from .analyzer import SpanAnalyzer, weighted_average


app = typer.Typer()


@app.command()
def main(
    input_path: Path = typer.Argument(
        ..., exists=True, dir_okay=False, help="Path to .spacy file"
    ),
    spacy_model: Optional[str] = typer.Option(
        None, help="Loadable spaCy pipeline (uses spacy.blank('en') if not provided)"
    ),
    verbose: bool = typer.Option(
        False, "--verbose", help="Show descriptions for each span property"
    ),
):
    nlp = spacy.load(spacy_model) if spacy_model else spacy.blank("en")

    # Read DocBin file
    doc_bin = DocBin().from_disk(input_path)
    docs: List[Doc] = list(doc_bin.get_docs(nlp.vocab))
    msg.info(f"Loaded {len(docs)} from {input_path}")

    # Perform span analysis
    analyzer = SpanAnalyzer(docs)
    msg.text(f"Spans Keys: {list(analyzer.keys)}")

    # Store in a variable because we will reuse this
    frequencies = analyzer.frequency

    msg_template(
        data=analyzer.frequency,
        divider="Span Type Frequency",
        header=("Span Type", "Frequency"),
        doc=SpanAnalyzer.frequency.__doc__,
        verbose=verbose,
    )

    msg_template(
        data=analyzer.length,
        divider="Span Type Length",
        header=("Span Type", "Length"),
        doc=SpanAnalyzer.length.__doc__,
        verbose=verbose,
        frequencies=frequencies,
    )

    msg_template(
        data=analyzer.span_distinctiveness,
        divider="Span Distinctiveness",
        header=("Span Key", "Span Distinctiveness"),
        doc=SpanAnalyzer.span_distinctiveness.__doc__,
        verbose=verbose,
        frequencies=frequencies,
    )

    msg_template(
        data=analyzer.boundary_distinctiveness,
        divider="Span Boundary Distinctiveness",
        header=("Span Key", "Boundary Distinctiveness"),
        doc=SpanAnalyzer.boundary_distinctiveness.__doc__,
        verbose=verbose,
        frequencies=frequencies,
    )


def msg_template(
    data: Dict[str, Any],
    divider: str,
    header: Tuple,
    doc: str,
    verbose: bool,
    frequencies: Optional[Dict[str, float]] = None,
):
    msg.divider(divider)
    if verbose:
        msg.text(cleandoc(doc))
    for spans_key, values in data.items():
        msg.text(f"In spans key: {spans_key}")
        msg.table(values, header=header, divider=True)

    if frequencies:
        # Compute weighted average
        w_avg = weighted_average(data, frequencies)
        msg.text("Weighted Average (by frequency)")
        msg.table(w_avg, header=("Spans Key", "Average"), divider=True)


if __name__ == "__main__":
    app()
